/*
	Copyright (C) 2011 Steve Thomas <steve AT tobtu DOT com>

	This file is part of MySQL323 Collider.

	MySQL323 Collider is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	MySQL323 Collider is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with MySQL323 Collider.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "mysql323thread.h"
#include "mysql323keyspace.h"
#include <stdio.h>

uint32_t search(uint32_t hash1, uint32_t hash2, passwordElement **miniLookupTable, uint64_t &foundPw);

MySQL323Thread::MySQL323Thread()
{
	m_queue = NULL;
	MUTEX_CREATE(m_mutex);
}

MySQL323Thread::~MySQL323Thread()
{
	MUTEX_DELETE(m_mutex);
}

int MySQL323Thread::stopRunning()
{
	MUTEX_LOCK(m_mutex);
	if (m_queue == NULL)
	{
		MUTEX_UNLOCK(m_mutex);
		return 1;
	}
	m_queue->flush();
	MUTEX_UNLOCK(m_mutex);
	return 0;
}

void *MySQL323Thread::debugrun(void *arg)
{
	return threadFunc(arg);
}

void *MySQL323Thread::threadFunc(void *arg)
{
	uint64_t start, pw;
	double time;
	WorkQueue *inQueue, *outQueue;
	mySQL323WorkQueueNode *node;
	Atomic *keySpacePos;
	passwordElement **indexedLookupTable;
	uint32_t *bitmap;
	uint8_t *prevHash;
	char *charSet;
	TIMER_TYPE t1, t2;
	uint32_t minCount, startCount, count, countDone, hash1, hash2, pwSuffixLen, firstNewChar, charSetLen, threadId;
	int ret;
	uint32_t i;
	uint32_t ch1, ch2;
	uint32_t state1, state1t, prevState1;
	uint32_t state2,          prevState2;
	uint32_t fullSum, prefixSum, sum;
	int32_t pos;
	uint32_t tmp, hashPos;
	char pwSuffix[MAX_LEN + 1], foundPw[MAX_LEN + 1];

	inQueue                         =   ((mySQL323ThreadArgs*)arg)->inQueue;
	outQueue                        =   ((mySQL323ThreadArgs*)arg)->outQueue;
	keySpacePos                     =   ((mySQL323ThreadArgs*)arg)->keySpacePos;
	MySQL323KeySpace keySpace       = *(((mySQL323ThreadArgs*)arg)->keySpace);
	MySQL323KeySpace prefixKeySpace = *(((mySQL323ThreadArgs*)arg)->prefixKeySpace);
	indexedLookupTable              =   ((mySQL323ThreadArgs*)arg)->indexedLookupTable;
	bitmap                          =   ((mySQL323ThreadArgs*)arg)->bitmap;
	prevHash                        =   ((mySQL323ThreadArgs*)arg)->prevHash;
	threadId                        =   ((mySQL323ThreadArgs*)arg)->threadId;
	if (keySpace.getMaxPwLen() + prefixKeySpace.getMaxPwLen() + 2 > MAX_LEN)
	{
		fprintf(stderr, "ERROR: **** Key space max password length (%u) > MAX_LEN (%u) ****\n", keySpace.getMaxPwLen() + prefixKeySpace.getMaxPwLen() + 2, MAX_LEN);
		return NULL;
	}
	MUTEX_LOCK(m_mutex);
	m_queue = inQueue;
	MUTEX_UNLOCK(m_mutex);

	TIMER_FUNC(t1);
	node = (mySQL323WorkQueueNode*) inQueue->pop();
	if (node != NULL)
	{
		startCount = 2000 / node->data->hashes.size();
		if (startCount < 256)
		{
			startCount = 256;
		}
	}
	while (node != NULL)
	{
		start = keySpacePos->add(startCount);
		for (hashPos = 0; hashPos < node->data->hashes.size(); hashPos++)
		{
			count = startCount;
			hash1 = node->data->hashes[hashPos].hash1;
			hash2 = node->data->hashes[hashPos].hash2;
			prefixSum = node->data->sum;
			pwSuffix[0] = 0;
			ret = keySpace.reset(start);
			if (ret == -1)
			{
				node->data->count = 0;
				node->data->time = 0;
				node->data->threadId = threadId;
				outQueue->push((workQueueNode*) node);
				break;
			}
			if (count <= (uint32_t) ret)
			{
				count = 1;
			}
			else
			{
				count -= (uint32_t) ret;
			}
			countDone = 0;
			do
			{
				if (keySpace.next(pwSuffix, pwSuffixLen, firstNewChar, &charSet, charSetLen))
				{
					break;
				}
				fullSum = prefixSum + 34 + 126;
				for (i = 0; i < pwSuffixLen; i++)
				{
					fullSum += (uint32_t) pwSuffix[i];
				}
				for (i = 0; i < charSetLen; i++)
				{
					sum = fullSum;
					// Reverse hash
					// Reverse state1
					tmp = ((uint32_t) prevHash[((hash1 & 63) << 11) | ((sum & 63) << 5) | (((uint32_t) charSet[i] & 63) >> 1)] + sum) * (uint32_t) charSet[i];
					prevState1 = hash1 ^ tmp;
					prevState1 = hash1 ^ ((prevState1 << 8) + tmp);
					prevState1 = hash1 ^ ((prevState1 << 8) + tmp);
					state1     = hash1 ^ ((prevState1 << 8) + tmp);
					// Reverse state2
					prevState2 = hash2 - state1;
					prevState2 = hash2 - ((prevState2 << 8) ^ state1);
					prevState2 = hash2 - ((prevState2 << 8) ^ state1);
					state2     = hash2 - ((prevState2 << 8) ^ state1);

					for (pos = pwSuffixLen - 1; pos >= 0; pos--)
					{
						sum -= (uint32_t) pwSuffix[pos];
						// Reverse state1
						tmp = ((uint32_t) prevHash[((state1 & 63) << 11) | ((sum & 63) << 5) | (((uint32_t) pwSuffix[pos] & 63) >> 1)] + sum) * (uint32_t) pwSuffix[pos];
						prevState1 = state1 ^ tmp;
						prevState1 = state1 ^ ((prevState1 << 8) + tmp);
						prevState1 = state1 ^ ((prevState1 << 8) + tmp);
						state1     = state1 ^ ((prevState1 << 8) + tmp);
						// Reverse state2
						prevState2 = state2 - state1;
						prevState2 = state2 - ((prevState2 << 8) ^ state1);
						prevState2 = state2 - ((prevState2 << 8) ^ state1);
						state2     = state2 - ((prevState2 << 8) ^ state1);
					}

					// Doing only 34+126, 36+124, 38+122, ..., 80+80, ..., 126+34 instead of any+any cuts over head in half
					// (126 - 34) / 2 + 1 = 47 characters
					// (47 - 1) * 2 + 1 = 93 different sums vs 1 sum
					// 47 * 47 = 2,209 passwords vs 47 passwords
					// 2209 / 97 vs 47 / 1
					// 23.7527 vs 47
					// 47 / 23.7527 = 1.9787
					ch1 = 34;
					ch2 = 126;
					sum -= ch1;
					while (ch1 <= 126)
					{
						// prefix + ch2 + ch1 + reversed
						// Reverse ch1
						tmp = ((uint32_t) prevHash[((state1 & 63) << 11) | ((sum & 63) << 5) | ((ch1 & 63) >> 1)] + sum) * ch1;
						prevState1 = state1 ^ tmp;
						prevState1 = state1 ^ ((prevState1 << 8) + tmp);
						prevState1 = state1 ^ ((prevState1 << 8) + tmp);
						state1t    = state1 ^ ((prevState1 << 8) + tmp);
						// Reverse ch2
						prevState2 = state2 - state1t;
						prevState2 = state2 - ((prevState2 << 8) ^ state1t);
						prevState2 = state2 - ((prevState2 << 8) ^ state1t);
						prevState2 = (state2 - ((prevState2 << 8) ^ state1t)) & 0x7fffffff;
						if (bitmap == NULL || (bitmap[prevState2 >> 5] & (1 << (prevState2 & 0x1f))) != 0)
						{
							tmp = ((uint32_t) prevHash[((state1t & 63) << 11) | ((prefixSum & 63) << 5) | ((ch2 & 63) >> 1)] + prefixSum) * ch2;
							prevState1 = state1t ^ tmp;
							prevState1 = state1t ^ ((prevState1 << 8) + tmp);
							prevState1 = state1t ^ ((prevState1 << 8) + tmp);
							prevState1 = (state1t ^ ((prevState1 << 8) + tmp)) & 0x7fffffff;
							if (search(prevState1, prevState2, indexedLookupTable, pw))
							{
								// pw -> foundPw
								pos = prefixKeySpace.get(pw, foundPw);
								foundPw[pos    ] = ch2;
								foundPw[pos + 1] = ch1;
								memcpy(foundPw + pos + 2, pwSuffix, pwSuffixLen);
								foundPw[pos + 2 + pwSuffixLen    ] = charSet[i];
								foundPw[pos + 2 + pwSuffixLen + 1] = 0;
#if (_MSC_VER == 1700) // Are you fucking serious
								node->data->foundPws.push_back(make_pair(hashPos, (string) foundPw));
#else
								node->data->foundPws.push_back(make_pair<uint32_t,string>(hashPos, (string) foundPw));
#endif
							}
						}
						ch1 += 2;
						ch2 -= 2;
						sum -= 2;
					}
				}
				countDone += charSetLen;
			} while (countDone < count);
		}
		node->data->threadId = threadId;
		node->data->count = countDone;
		TIMER_FUNC(t2);
		time = TIMER_DIFF(t1, t2);
		node->data->time = time;
		t1 = t2;

		// Adjust startCount
		if (time < 0.080 || time > 0.125)
		{
			if (time <= 0.015)
			{
				startCount = 4 * countDone;
			}
			else if (time < 0.080)
			{
				startCount = (uint32_t) (countDone * 0.1 / time);
			}
			else
			{
				count = (uint32_t) (countDone * 0.1 / time);
				startCount = (uint32_t) (countDone / 1.5);
				if (startCount < count)
				{
					startCount = count;
				}
			}
			minCount = 2000 / node->data->hashes.size();
			if (startCount < minCount)
			{
				startCount = minCount;
			}
			if (startCount < 256)
			{
				startCount = 256;
			}
		}

		outQueue->push((workQueueNode*) node);
		node = (mySQL323WorkQueueNode*) inQueue->pop();
	}
	MUTEX_LOCK(m_mutex);
	m_queue = NULL;
	MUTEX_UNLOCK(m_mutex);
	return NULL;
}

uint32_t search(uint32_t hash1, uint32_t hash2, passwordElement **miniLookupTable, uint64_t &foundPw)
{
	uint64_t hash;
	passwordElement *lookupTable;
	int32_t lo, hi, mid;

	hash = (((uint64_t) hash2) << 32) | hash1;
	lookupTable = miniLookupTable[hash2 >> 7];
	lo = 0;
	hi = miniLookupTable[(hash2 >> 7) + 1] - lookupTable;
	while (lo <= hi)
	{
		mid = (lo + hi) / 2;
		if (hash < lookupTable[mid].hash)
		{
			hi = mid - 1;
		}
		else if (hash > lookupTable[mid].hash)
		{
			lo = mid + 1;
		}
		else
		{
			foundPw = lookupTable[mid].pw;
			return 1;
		}
	}
	return 0;
}
